package org.jenkinsci.plugins.p4_client.client;

import hudson.model.TaskListener;
import hudson.security.ACL;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import jenkins.model.Jenkins;

import org.acegisecurity.Authentication;
import org.jenkinsci.plugins.p4_client.credentials.P4StandardCredentials;
import org.jenkinsci.plugins.p4_client.workspace.ManualWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.StaticWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.StreamWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.TemplateWorkspaceImpl;
import org.jenkinsci.plugins.p4_client.workspace.Workspace;

import com.cloudbees.plugins.credentials.CredentialsProvider;
import com.cloudbees.plugins.credentials.domains.DomainRequirement;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.client.IClientSummary.ClientLineEnd;
import com.perforce.p4java.core.file.FileAction;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.AccessException;
import com.perforce.p4java.exception.ConnectionException;
import com.perforce.p4java.exception.RequestException;
import com.perforce.p4java.impl.generic.client.ClientOptions;
import com.perforce.p4java.impl.generic.client.ClientView;
import com.perforce.p4java.impl.generic.client.ClientView.ClientViewMapping;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.option.client.ReconcileFilesOptions;
import com.perforce.p4java.option.client.RevertFilesOptions;
import com.perforce.p4java.option.client.SyncOptions;
import com.perforce.p4java.server.IOptionsServer;

public class ConnectionHelper {

	private static Logger logger = Logger.getLogger(ConnectionHelper.class
			.getName());

	private final ConnectionConfig connectionConfig;
	private final AuthorisationConfig authorisationConfig;
	private final IOptionsServer connection;
	private final TaskListener listener;
	private IClient iclient;

	public ConnectionHelper(String credentialID, TaskListener listener)
			throws Exception {
		this.listener = listener;
		P4StandardCredentials credential = findCredential(credentialID);
		this.connectionConfig = new ConnectionConfig(credential);
		this.authorisationConfig = new AuthorisationConfig(credential);
		this.connection = ConnectionFactory.getConnection(connectionConfig);
	}

	public ConnectionHelper(P4StandardCredentials credential) throws Exception {
		this.listener = null;
		this.connectionConfig = new ConnectionConfig(credential);
		this.authorisationConfig = new AuthorisationConfig(credential);
		this.connection = ConnectionFactory.getConnection(connectionConfig);
	}

	public boolean isConnected() {
		return connection.isConnected();
	}

	public boolean isUnicode() {
		try {
			return connection.getServerInfo().isUnicodeEnabled();
		} catch (Exception e) {
			return false;
		}
	}

	public boolean login() throws Exception {
		connection.setUserName(authorisationConfig.getUsername());

		// CHARSET is not defined (only for client access)
		if (connection.getServerInfo().isUnicodeEnabled()) {
			connection.setCharsetName("utf8");
		}

		// exit early if logged in
		if (isLogin())
			return true;

		switch (authorisationConfig.getType()) {
		case PASSWORD:
			String pass = authorisationConfig.getPassword();
			connection.login(pass);
			break;

		case TICKET:
			String ticket = authorisationConfig.getTicketValue();
			connection.setAuthTicket(ticket);
			break;

		case TICKETPATH:
			String path = authorisationConfig.getTicketPath();
			connection.setTicketsFilePath(path);
			break;

		default:
			throw new Exception("Unknown Authorisation type: "
					+ authorisationConfig.getType());
		}

		return isLogin();
	}

	public void logout() throws Exception {
		if (isLogin()) {
			connection.logout();
		}
	}

	private boolean isLogin() throws Exception {
		String status = connection.getLoginStatus();
		if (status.contains("not necessary")) {
			return true;
		}
		if (status.contains("ticket expires in")) {
			return true;
		}
		// If there is a broker or something else that swallows the message
		if (status.isEmpty()) {
			return true;
		}

		logger.info("P4:login failed '" + status + "'");
		return false;
	}

	/**
	 * Create default client from Configuration
	 * 
	 * @param user
	 * @return
	 * @throws Exception
	 */
	public void setClient(Workspace workspace) throws Exception {

		if (connectionConfig.isUnicode()) {
			String charset = "utf8";
			connection.setCharsetName(charset);
		}

		switch (workspace.getType()) {
		case MANUAL:
			ManualWorkspaceImpl manualWs = (ManualWorkspaceImpl) workspace;
			iclient = connection.getClient(manualWs.getName());
			if (iclient == null) {
				logger.info("... Creating manual client: " + manualWs.getName());
				Client implClient = new Client();
				implClient.setName(manualWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(manualWs.getName());
			}

			String user = authorisationConfig.getUsername();
			iclient.setOwnerName(user);

			ClientOptions options = new ClientOptions();
			options.setAllWrite(manualWs.getSpec().allwrite);
			options.setClobber(manualWs.getSpec().clobber);
			options.setCompress(manualWs.getSpec().compress);
			options.setLocked(manualWs.getSpec().locked);
			options.setModtime(manualWs.getSpec().modtime);
			options.setRmdir(manualWs.getSpec().rmdir);
			iclient.setOptions(options);

			iclient.setStream(manualWs.getSpec().getStream());
			iclient.setLineEnd(parseLineEnd(manualWs.getSpec().getLine()));

			ClientView clientView = new ClientView();
			int order = 0;
			for (String line : manualWs.getSpec().getView().split("\\n")) {
				ClientViewMapping entry = new ClientViewMapping(order, line);
				order++;
				clientView.addEntry(entry);
			}
			iclient.setClientView(clientView);

			break;

		case STATIC:
			StaticWorkspaceImpl staticWs = (StaticWorkspaceImpl) workspace;
			iclient = connection.getClient(staticWs.getName());
			if (iclient == null) {
				throw new Exception("Undefined workspace: "
						+ staticWs.getName());
			}
			break;

		case TEMPLATED:
			TemplateWorkspaceImpl templateWs = (TemplateWorkspaceImpl) workspace;
			String template = templateWs.getTemplateName();
			templateWs.getName();
			iclient = connection.getClient(templateWs.getName());
			if (iclient == null) {
				logger.info("Creating template client: " + templateWs.getName());
				Client implClient = new Client();
				implClient.setName(templateWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(templateWs.getName());
			}
			connection.switchClientView(template, templateWs.getName(), null);
			break;

		case STREAM:
			StreamWorkspaceImpl streamWs = (StreamWorkspaceImpl) workspace;
			if (iclient == null) {
				logger.info("Creating stream client: " + streamWs.getName());
				Client implClient = new Client();
				implClient.setName(streamWs.getName());
				connection.createClient(implClient);
				iclient = connection.getClient(streamWs.getName());
			}
			iclient.setStream(streamWs.getStreamName());
			break;

		default:
			throw new Exception("Unknown workspace type: "
					+ workspace.getType());
		}

		// Ensure root and host fields are set
		iclient.setRoot(workspace.getRootPath());
		iclient.setHostName(workspace.getHostName());

		// Set clobber on to ensure workspace is always good
		ClientOptions options = new ClientOptions();
		options.setClobber(true);
		iclient.setOptions(options);

		// Save client spec
		iclient.update();

		// Set active client for this connection
		connection.setCurrentClient(iclient);
	}
	
	public void deleteClient(Workspace workspace) throws Exception {
		if (connectionConfig.isUnicode()) {
			String charset = "utf8";
			connection.setCharsetName(charset);
		}

		String name = workspace.getName();
		connection.deleteClient(name, true);
	}

	/**
	 * Test to see if workspace is at the latest revision.
	 * 
	 * @throws Exception
	 */
	public boolean updateFiles() throws Exception {
		// build file revision spec
		List<IFileSpec> syncFiles;
		syncFiles = FileSpecBuilder.makeFileSpecList("//...");

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		syncOpts.setNoUpdate(true);
		List<IFileSpec> syncMsg = iclient.sync(syncFiles, syncOpts);

		for (IFileSpec fileSpec : syncMsg) {
			if (fileSpec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();
				if (msg.contains("file(s) up-to-date.")) {
					return false;
				}
			}
		}
		return true;
	}
	
	// TODO for Jenkins 'Recent Changes' Report feature
	public int lastChange() throws Exception {
		Map<String, Object>[] map = connection.execMapCmd("cstat", new String[] {"//..."}, null);
		logger.info(map.toString());
		return 0;
	}

	/**
	 * Sync latest revision of files to workspace.
	 * 
	 * @throws Exception
	 */
	public boolean syncFiles() throws Exception {
		boolean success = true;

		log("SCM Task: syncing files at HEAD");

		// build file revision spec
		List<IFileSpec> syncFiles;
		syncFiles = FileSpecBuilder.makeFileSpecList("//...");

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		log("... sync //...");
		List<IFileSpec> syncMsg = iclient.sync(syncFiles, syncOpts);
		success &= validateFileSpecs(syncMsg, "file(s) up-to-date.");

		return success;
	}

	/**
	 * Sync files to workspace at the specified change.
	 * 
	 * @param change
	 * @throws Exception
	 */
	public boolean syncFiles(int change) throws Exception {
		boolean success = true;

		log("SCM Task: syncing files at change: " + change);

		// build file revision spec
		List<IFileSpec> files;
		String revisions = "//...@" + change;
		files = FileSpecBuilder.makeFileSpecList(revisions);

		// Sync revision to re-edit
		SyncOptions syncOpts = new SyncOptions();
		log("... sync " + revisions);
		List<IFileSpec> syncMsg = iclient.sync(files, syncOpts);
		success &= validateFileSpecs(syncMsg, "file(s) up-to-date.");

		return success;
	}

	/**
	 * Cleans up the Perforce workspace after a previous build. Removes all
	 * pending and abandoned files (equivalent to 'p4 revert -w').
	 * 
	 * @throws Exception
	 */
	public boolean tidyWorkspace() throws Exception {
		boolean success = true;

		log("SCM Task: cleanup workspace: " + iclient.getName());

		// relies on workspace view for scope.
		List<IFileSpec> files;
		files = FileSpecBuilder.makeFileSpecList("//...");

		// revert all pending and shelved revisions
		RevertFilesOptions rOpts = new RevertFilesOptions();
		log("... [list] = revert //...");
		List<IFileSpec> list = iclient.revertFiles(files, rOpts);
		success &= validateFileSpecs(list, "not opened on this client");

		// check for added files and remove...
		log("... rm [list] | ABANDONED");
		for (IFileSpec file : list) {
			if (file.getAction() == FileAction.ABANDONED) {
				String path = depotToLocal(file);
				if (path != null) {
					File unlink = new File(path);
					unlink.delete();
				}
			}
		}

		// check status - find all extra files
		ReconcileFilesOptions statusOpts = new ReconcileFilesOptions();
		statusOpts.setOutsideAdd(true);
		statusOpts.setNoUpdate(true);
		log("... [list] = reconcile -n -a //...");
		List<IFileSpec> extra = iclient.reconcileFiles(files, statusOpts);
		success &= validateFileSpecs(extra, "- no file(s) to reconcile");

		// remove added files
		log("... rm [list]");
		for (IFileSpec e : extra) {
			String path = depotToLocal(e);
			if (path != null) {
				File unlink = new File(path);
				unlink.delete();
			}
		}

		// check status - find all missing or changed
		statusOpts = new ReconcileFilesOptions();
		statusOpts.setNoUpdate(true);
		log("... [list] = reconcile -n //...");
		List<IFileSpec> update = iclient.reconcileFiles(files, statusOpts);
		success &= validateFileSpecs(update, "also opened by",
				"no file(s) to reconcile", "must sync/resolve");

		// force sync to update files
		SyncOptions syncOpts = new SyncOptions();
		syncOpts.setForceUpdate(true);
		log("... sync -f [list]");
		List<IFileSpec> syncMsg = iclient.sync(update, syncOpts);
		success &= validateFileSpecs(syncMsg, "file(s) up-to-date.",
				"file does not exist");

		return success;
	}

	/**
	 * Workaround for p4java bug. The 'setLocalSyntax(true)' option does not
	 * provide local syntax, so I have to use 'p4 where' to translate through
	 * the client view.
	 * 
	 * @param fileSpec
	 * @return
	 * @throws Exception
	 */
	private String depotToLocal(IFileSpec fileSpec) throws Exception {
		String depotPath = fileSpec.getDepotPathString();
		if (depotPath == null)
			return null;

		List<IFileSpec> dSpec = FileSpecBuilder.makeFileSpecList(depotPath);
		List<IFileSpec> lSpec = iclient.where(dSpec);
		String path = lSpec.get(0).getLocalPathString();
		return path;
	}

	/**
	 * Unshelve review into workspace. Workspace is sync'ed to head first then
	 * review unshelved.
	 * 
	 * @param review
	 * @throws Exception
	 */
	public boolean unshelveFiles(int review) throws Exception {
		boolean success = true;

		log("SCM Task: unshelve review: " + review);

		// build file revision spec
		List<IFileSpec> files;
		files = FileSpecBuilder.makeFileSpecList("//...");

		// Sync workspace to head
		SyncOptions syncOpts = new SyncOptions();
		log("... sync //...");
		List<IFileSpec> syncMsg = iclient.sync(files, syncOpts);
		success &= validateFileSpecs(syncMsg, "file(s) up-to-date.");

		// Unshelve change for review
		List<IFileSpec> shelveMsg;
		log("... unshelve -f -s " + review + " //...");
		shelveMsg = iclient.unshelveChangelist(review, files, 0, true, false);
		success &= validateFileSpecs(shelveMsg, "also opened by",
				"no such file(s)");

		// Remove opened files from have list.
		RevertFilesOptions rOpts = new RevertFilesOptions();
		rOpts.setNoUpdate(true);
		log("... revert -k //...");
		List<IFileSpec> rvtMsg = iclient.revertFiles(files, rOpts);
		success &= validateFileSpecs(rvtMsg,
				"file(s) not opened on this client");

		return success;
	}

	/**
	 * Disconnect from the Perforce Server.
	 * 
	 * @throws Exception
	 */
	public void disconnect() throws Exception {
		connection.disconnect();
	}

	/**
	 * Look for a message in the returned FileSpec from operation.
	 * 
	 * @param fileSpecs
	 * @param ignore
	 * @return
	 * @throws ConverterException
	 */
	public boolean validateFileSpecs(List<IFileSpec> fileSpecs,
			String... ignore) throws Exception {
		for (IFileSpec fileSpec : fileSpecs) {
			if (fileSpec.getOpStatus() != FileSpecOpStatus.VALID) {
				String msg = fileSpec.getStatusMessage();

				// superfluous p4java message
				boolean unknownMsg = true;
				ArrayList<String> ignoreList = new ArrayList<String>();
				ignoreList.addAll(Arrays.asList(ignore));
				for (String istring : ignoreList) {
					if (msg.contains(istring)) {
						// its a known message
						unknownMsg = false;
					}
				}

				// check and report unknown message
				if (unknownMsg) {
					log("P4JAVA: " + msg);
					logger.warning("p4java: " + msg);
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Finds a Perforce Credential based on the String id.
	 * 
	 * @return a P4StandardCredentials credential or null if not found.
	 */
	private static P4StandardCredentials findCredential(String id) {
		Class<P4StandardCredentials> type = P4StandardCredentials.class;
		Jenkins scope = Jenkins.getInstance();
		Authentication acl = ACL.SYSTEM;
		DomainRequirement domain = new DomainRequirement();

		List<P4StandardCredentials> list;
		list = CredentialsProvider.lookupCredentials(type, scope, acl, domain);

		for (P4StandardCredentials c : list) {
			if (c.getId().equals(id)) {
				return c;
			}
		}
		return null;
	}

	private ClientLineEnd parseLineEnd(String line) {
		for (ClientLineEnd end : ClientLineEnd.values()) {
			end.name().equalsIgnoreCase(line);
			return end;
		}
		return ClientLineEnd.LOCAL;
	}

	public void log(String msg) {
		if (listener == null) {
			return;
		}

		listener.getLogger().println(msg);
	}
}
